/*
 * ggblocks.c - functions and structures to store a game board ('WORLD') and 
 *      compute and apply moves in it. 
 *
 *    Copyright (C) 2003 Eric Cheung.
 *
 *    This file is part of GG's Blocks.
 *
 *    GG's Blocks is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    GG's Blocks is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with GG's Blocks; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * ===========================================================================
 *  This code should be fairly self-explanatory. The only point I want to give
 *  more details on is regarding world.emptyCol, world.topUpdate, and
 *  world.bottomUpdate.
 *
 *  Each of these is there to make redrawing the world faster, so we (almost)
 *  only redraw the squares that need to be redrawn.
 *
 *  "world.emptyCol" - the number of empty columns on the LEFT of the playfield
 *    eg.
 *     01234567
 *    ==========
 *   0|    R FF|
 *   1|    R FF|
 *   2|    ggXg|
 *   3|   ggXXg|
 *   4|   ggggg|
 *    ==========
 *    for this playfield emptyCol == 3. This value is used when compacting the 
 *    world to speed up the compaction slightly.
 *
 *  "world.topUpdate[C]" and "world.bottomUpdate[C]" - the top and bottom 
 *    modified range for column number 'C'. A modified range means that inside
 *    this interval, the tiles have changed since the last redraw
 *    eg. suppose from the last field the player clicked on the X's so they
 *    became empties
 *     01234567
 *    ==========
 *   0|    R FF|
 *   1|    R FF|
 *   2|    gg g|
 *   3|   gg  g|
 *   4|   ggggg|
 *    ==========
 *    Then we will have
 *      w->topUpdate[5] = 3; w->bottomUpdate[5] = 3;
 *      w->topUpdate[6] = 2; w->bottomUpdate[6] = 3;
 *    and for every other column 'C'={0,1,2,3,4,7}
 *      w->topUpdate[C] = 4;  
 *      w->bottomUpdate[C] = 0;  
 *    
 *  When we redraw the screen in displayWorld, we only redraw the ranges that 
 *  have been marked as modified by topUpdate and bottomUpdate. After redrawing
 *  we reset topUpdate and bottomUpdate to mark everything as unmodified.
 * ===========================================================================
 */

#include <PalmOS.h>
#include <MemoryMgr.h>
#include <SysUtils.h>       // for SysRandom()
#include <Window.h>

#include "ggblocks.h"

// internal functions
static COORD getOffset(Int16 dir, COORD curr);
// returns true if there are still valid moves left
static Boolean worldMovesLeft(WORLD *w);
// mixes the world for a 5x5 area around the given coords
static void worldMix(WORLD *w, Int16 c, Int16 r);

// marks the entire world for redraw
void worldMarkModified(WORLD *w){
    Int16 col;

    for(col=0;col<WORLDCOL;col++){
        // set these counters to specify that the whole playfield should be
        // redrawn
        w->topUpdate[col] = 0;
        w->bottomUpdate[col] = WORLDROW-1;
    }
}

// random world
void worldGenerate(WORLD *w) {
    Int16 col,row;
    Int16 numMixer;

    w->emptyCol = 0;
    w->movesLeft = true;
    for(col=0;col<WORLDCOL;col++)
        for(row=0;row<WORLDROW;row++)
            w->tile[col][row] = (TILE)(SysRandom(0)%NUMDIFFTILES);

    // put mixer squares in
    numMixer = (SysRandom(0)%3)+2;
    while(numMixer){
        col = SysRandom(0)%WORLDCOL;
        row = SysRandom(0)%WORLDROW;

        // make sure we don't put two mixers on the same square
        if(w->tile[col][row] != MIXER){
            w->tile[col][row] = MIXER;
            numMixer--;
        }
    }

    // mark the whole world for redraw
    worldMarkModified(w);
}

// wipes out the move in the world. uses a BFS
Int32 worldApplyMove(WORLD *w, Int16 c, Int16 r) {
    COORD queue[WORLDCOL*WORLDROW]; 
    Boolean visited[WORLDCOL][WORLDROW];
    Int16 head = 0;
    Int16 tail = 0;
    Boolean legal;
    COORD tmp, curr;
    TILE moveTile;
    Int16 i;
    Int16 numRemoved;

    curr.c = c;
    curr.r = r;

    legal = false;

    // check this move is legal
    if(c<0 || c>=WORLDCOL || r<0 || r>=WORLDROW)
        return -1;

    moveTile = w->tile[c][r];       // the tile type of this move
    if(moveTile == EMPTY)
        return -1;

    if(moveTile == MIXER){
        // turn the mixer into a random tile
        w->tile[c][r] = SysRandom(0)%NUMDIFFTILES;

        // mix it
        worldMix(w,c,r);
        return 0;
    }

    for(i=0;i<4;i++){
        tmp = getOffset(i, curr);

        // don't do if we're out of bounds
        if(tmp.r<0 || tmp.r>=WORLDROW || tmp.c<0 || tmp.c>=WORLDCOL)
            continue;

        if(w->tile[tmp.c][tmp.r] == moveTile){
            // we're neighboured by at least one tile of the same
            // colour as ourselves
            legal = true;
            break;
        }
    }

    if(!legal)
        return -1;

    MemSet(visited, sizeof(Boolean)*WORLDCOL*WORLDROW, false);

    // push starting square onto queue
    queue[tail++] = curr;
    visited[curr.c][curr.r] = true;
    numRemoved = 0;

    while(head != tail) {
        // pop the top off the queue
        curr = queue[head++];

        // make a hole here
        numRemoved++;
        w->tile[curr.c][curr.r] = EMPTY; 
        // mark this tile for redraw
        if(curr.r < w->topUpdate[curr.c])
            w->topUpdate[curr.c] = curr.r;
        if(curr.r > w->bottomUpdate[curr.c])
            w->bottomUpdate[curr.c] = curr.r;

        // put our neighbours onto queue
        for(i=0;i<4;i++){
            // get the offsets of the 4 directions
            tmp = getOffset(i, curr);
                
            // don't do if we're out of bounds
            if(tmp.r<0 || tmp.r>=WORLDROW || tmp.c<0 || tmp.c>=WORLDCOL)
                continue;

            // don't do if we're already visited
            if(visited[tmp.c][tmp.r])
                continue;

            // don't do if we're a different type of tile
            if(w->tile[tmp.c][tmp.r] != moveTile)
                continue;

            // push onto queue
            queue[tail++] = tmp;
            // mark it as visited
            visited[tmp.c][tmp.r] = true;
        }
    }
   
    return (Int32)numRemoved*numRemoved*5;
}

// squeezes EMPTY squares out of the gameboard
void worldCompact(WORLD *w) {
    COORD tmp;
    Int16 i, col;

    // for each column, move downwards and check for holes; if any are found,
    // remove them
    for(tmp.c=w->emptyCol;tmp.c<WORLDCOL;tmp.c++) {
        TILE *currCol = w->tile[tmp.c];
        Int16 startRow;

        // skip empty space at top of this column
        for(startRow=0;currCol[startRow]==EMPTY && startRow<WORLDROW;
                        startRow++);

        // condense any holes we find in the remainder of this column
        for(tmp.r=startRow;tmp.r<WORLDROW;tmp.r++)
            if(currCol[tmp.r] == EMPTY){
                MemMove(&currCol[startRow+1], &currCol[startRow], (tmp.r-
                                        startRow)*sizeof(TILE));

                currCol[startRow] = EMPTY;

                // mark this section as modified
                if(startRow < w->topUpdate[tmp.c])
                    w->topUpdate[tmp.c] = startRow;
                if(tmp.r > w->bottomUpdate[tmp.c])
                    w->bottomUpdate[tmp.c] = tmp.r;
            }
    }

    // check for empty columns from the left; if any, remove them
    for(i=w->emptyCol;i<WORLDCOL;i++) {
        if(w->tile[i][WORLDROW-1] == EMPTY) {
            // shuffle across
            for(col=i-1;col>=0;col--) {
                // copy col i to i+1
                MemMove(&w->tile[col+1], &w->tile[col], WORLDROW*sizeof(TILE));

                // mark this whole column for redraw
                w->topUpdate[col+1] = 0;
                w->bottomUpdate[col+1] = WORLDROW-1;
            }

            MemSet(&w->tile[w->emptyCol], WORLDROW*sizeof(TILE), EMPTY);

            // mark this first column for redraw too
            w->topUpdate[w->emptyCol] = 0;
            w->bottomUpdate[w->emptyCol] = WORLDROW-1;

            w->emptyCol++;
        }
    }

    // update whether there are moves left or not
    w->movesLeft = worldMovesLeft(w);
}

// returns the points this world is worth if it were the finish state
Int32 worldFinalPointValue(WORLD *w) {
    Int16 percentComplete;
    Int16 numTile;
    Int16 i;
    Int16 r, c;

    // count the number of tiles remaining on the field
    for(c=numTile=0;c<WORLDCOL;c++)
        for(r=0;r<WORLDROW;r++)
            numTile += w->tile[c][r]==EMPTY?0:1;

    percentComplete = (WORLDROW*WORLDCOL-numTile)*100/(WORLDROW*WORLDCOL);
    i = percentComplete < 85? 0 : (percentComplete-85);

    return 25*i*i;
}

// Internal functions
// returns true if there are still valid moves left
static Boolean worldMovesLeft(WORLD *w) {
    TILE moveTile;
    Int16 i;
    COORD tmp, curr;

    // if any tile is neighboured by tiles of the same type, we have a move
    // left
    for(curr.c=0;curr.c<WORLDCOL;curr.c++)
        for(curr.r=0;curr.r<WORLDROW;curr.r++){
            moveTile = w->tile[curr.c][curr.r];

            if(moveTile == EMPTY)
                continue;
            if(moveTile == MIXER)
                return true;

            for(i=0;i<4;i++){
                tmp = getOffset(i, curr);

                // don't do if we're out of bounds
                if(tmp.r<0 || tmp.r>=WORLDROW || tmp.c<0 || tmp.c>=WORLDCOL)
                    continue;

                if(w->tile[tmp.c][tmp.r] == moveTile)
                    // we're neighboured by at least one tile of the same
                    // colour as ourselves
                    return true;
            }
        }

    return false;
}

static COORD getOffset(Int16 dir, COORD curr) {
    COORD tmp;

    switch(dir) {
    case 0:
        tmp.c = curr.c;
        tmp.r = curr.r-1;
        break;

    case 1:
        tmp.c = curr.c;
        tmp.r = curr.r+1;
        break;

    case 2:
        tmp.c = curr.c-1;
        tmp.r = curr.r;
        break;

    case 3:
        tmp.c = curr.c+1;
        tmp.r = curr.r;
        break;
    }

    return tmp;
}

// mixes the world for a 5x5 area around the given coords
static void worldMix(WORLD *w, Int16 c, Int16 r) {
    COORD queue[25];
    Int16 numQueue = 0;
    COORD tmp;
    Int16 i,j;
    TILE tileTmp;

    // get all the squares around here that are not EMPTY
    for(tmp.c=c-2;tmp.c<=c+2;tmp.c++)
        for(tmp.r=r-2;tmp.r<=r+2;tmp.r++){
            if(tmp.r<0 || tmp.r>=WORLDROW || tmp.c<0 || tmp.c>=WORLDCOL)
                continue;
            if(w->tile[tmp.c][tmp.r]!=EMPTY){
                queue[numQueue++] = tmp;

                // mark this tile for redraw
                if(tmp.r < w->topUpdate[tmp.c])
                    w->topUpdate[tmp.c] = tmp.r;
                if(tmp.r > w->bottomUpdate[tmp.c])
                    w->bottomUpdate[tmp.c] = tmp.r;
            }
        }
    
    // for each non-empty tile, swap it with another random non-empty
    // tile
    for(i=0;i<numQueue;i++){
        j = SysRandom(0)%numQueue;
        tileTmp = w->tile[queue[j].c][queue[j].r];
        w->tile[queue[j].c][queue[j].r] = w->tile[queue[i].c][queue[i].r];
        w->tile[queue[i].c][queue[i].r] = tileTmp;
    }
}
